/*
 * virdomaingroup.c: ExaData Domain Groups XML processing
 *
 * Copyright (c) 2019 Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <unistd.h>
#include <fcntl.h>

#include "domain_audit.h"
#include "internal.h"
#include "datatypes.h"
#include "viralloc.h"
#include "domain_conf.h"
#include "virlog.h"
#include "virerror.h"
#include "viraudit.h"
#include "virstring.h"
#include "virfile.h"
#include "virdomaingroup.h"

VIR_LOG_INIT("conf.domain_audit");

#define VIR_FROM_THIS VIR_FROM_NWFILTER

#if defined(ENABLE_EXADATA)

/* virDomainGroupBasePin
 *
 * virDomainGroupBasePtr dgbase - Domain Groups Base
 *
 * Arrange and acquire the Domain Group 'pcpus' (pCPU pairs) cores per
 * QEMU 'cpuguestmask' available content and place them in the Domain
 * Group allocated 'pcpumask' will start acting as the 'cpuguestmask'
 * for this Domain Group from now on. */
int
virDomainGroupBasePin(virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virBitmapPtr pcpumask = NULL;
    virBitmapPtr cpuguestmask = dgbase->cpuguestmask;
    virCapsHostPtr host = dgbase->host;
    virDomainGroupDefPtr *dgrp = dgbase->dgroups;
    virCapsHostNUMACellPtr *numaCell = NULL;
    size_t nnumaCell, cores, threads;

    /* If we ever end up here without host capabilities or a host->cpu
     * description of the pCPUs it is pointless to render the listed
     * domain groups */
    if (!host || !host->cpu) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Failed to render the defined domain groups because"
                         " the required host capabilities are missing"));
        goto error;
    }

    numaCell = host->numaCell;
    nnumaCell = host->nnumaCell;
    cores = host->cpu->cores;
    threads = host->cpu->threads;

    while (*dgrp) {
        bool FillHost = false;
        size_t pcores, cell = 0;
        ssize_t pcpu = -1;
        virDomainGroupDefPtr curr = *dgrp++;

        /* Proceed if this Domain Group is not in use. */
        if (!curr->DG_usecnt)
            continue;

        /* If there's no change of pcpus to this DG then
         * there is no need to make any change either. */
        if (curr->pcpus == curr->DG_prsvd)
            continue;

        /* Allocate the Domain Group pCPU bitmap */
        if (!(pcpumask = virBitmapNew(virBitmapSize(cpuguestmask) * curr->oscnt)))
            goto error;

        /* Despite waste we need round to physical cores */
        pcores = (curr->pcpus / threads) + ((curr->pcpus % threads) ? 1 : 0);

        /* Traverse the set of physical cores given for this Domain Group (pCPUs). */
        do {

            /* Assure the handed physical core selects a NUMA node within range. */
            if (cell < nnumaCell) {
                size_t i, tpc = threads;

                /* Select the host NUMA node to start from; if this
                 * concerns NUMA node #0 and there are no sufficient physical
                 * cores available under 'pcores' to fill all the host NUMA
                 * nodes, start seeking a NUMA node start from that holds
                 * the first available core. */
                if (FillHost || ((cell == 0) && ((pcores / nnumaCell) == 0))) {
                    size_t ncell = 0, target_pcpu, first_pcpu;

                    /* Arm FillHost */
                    FillHost = true;

                    /* Maximize 'target' and 'first' pCPU. */
                    target_pcpu = first_pcpu = dgbase->dgroup_guestpcpus;

                    /* Walk all the host NUMA nodes. */
                    for (cell = 0; cell < nnumaCell; cell++) {
                        pcpu = -1;

                        /* Walk all pCPU (pairs) under this NUMA node. */
                        for (i = 0; i < numaCell[cell]->ncpus; i++) {
                            if ((pcpu = virBitmapNextSetBit(numaCell[cell]->cpus[i].siblings, pcpu)) >= 0) {

                                /* If 'cpuguestmask' indicates a fit,
                                 * then remember per 'first_pcpu' NUMA node
                                 * relative number. */
                                if (virBitmapIsBitSet(cpuguestmask, pcpu)) {
                                    first_pcpu = pcpu - (cell * cores);
                                    break;
                                }
                            }
                        }

                        /* Sort the lowest NUMA node found fit and mark
                         * its NUMA node#. */
                        if (target_pcpu > first_pcpu) {
                            target_pcpu = first_pcpu;
                            ncell = cell;
                        }
                    }

                    /* Supersede the NUMA node start point for core
                     * placement before acquiring the next NUMA node
                     * sibling. */
                    cell = ncell;
                }

                /* Circle all sibling threads for this core */
                for (i = 0; i < numaCell[cell]->ncpus && tpc; i++) {
                    size_t thr = threads;

                    pcpu = -1;
                    while (thr--) {

                        if ((pcpu = virBitmapNextSetBit(numaCell[cell]->cpus[i].siblings, pcpu)) >= 0) {

                            if (virBitmapIsBitSet(cpuguestmask, pcpu)) {
                                if (virBitmapSetBitExpand(pcpumask, pcpu) < 0) {
                                    virBitmapFree(pcpumask);
                                    goto error;
                                }

                                /* Remove the 'bit' from the guest available
                                 * map (cpuguestmask) and place it under the DG
                                 * allocated map (pcpumask).  Further increment
                                 * the pCPU reserved count.
                                 * Up on suc6 curr->pcpus and curr->DG_prsvd
                                 * should be equal. */
                                ignore_value(virBitmapClearBit(cpuguestmask, pcpu));
                                ignore_value(virBitmapSetBit(pcpumask, pcpu));
                                curr->DG_prsvd++;

                                tpc--;
                            }
                        }
                    }
                }

                /* Get to the next NUMA node */
                if (++cell >= nnumaCell)
                    cell = 0;
            }
        } while (--pcores);

        /* If we failed making the pCPU domain group allocation push
         * back all allocated pCPUs into the cpuguestmask and free the
         * intermediate allocated pcpumask. */
        if (curr->pcpus != curr->DG_prsvd) {

            pcpu = -1;
            while ((pcpu = virBitmapNextSetBit(pcpumask, pcpu)) >= 0) {
                ignore_value(virBitmapSetBit(cpuguestmask, pcpu));
                curr->DG_prsvd--;
            }

            virBitmapFree(pcpumask);

            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Failed to allocate %zu pCPUs"
                             " for domain group '%s'"),
                           curr->pcpus, curr->name);
            goto error;
        }

        /* next */
        curr->pcpumask = pcpumask;
        curr->DG_id = ++dgbase->dgroup_id;
    }

    ret = 0;

 error:
    return ret;
}


/**
 * virDomainGroupAccumulateCPUs:
 *
 * Traverse all Domain Group subscribed guest Domains and
 * under the condition of CPU Packing have the vCPU:pCPU
 * pinning per guest Domain vCPU removed from cpuguestmask
 * or re-claimed to cpuguestmask.
 *
 * Parameters:
 *   dgrp:  domain group for which we perform this request.
 *
 *   claim:
 *          false - remove the vCPU from the cpuguestmask
 *          true  - restore the vCPU under the cpuguestmask and have
 *                  the specific vCPU Packing undone.
 *
 *   refresh:
 *          false - rebuild the cpuguestmask only.
 *          true  - if called from crash recoverey or service restart
 *                  need to rebuild all statistics from scratch.
 *
 * Adjust the Domain Group oversubscription indicator.
 */
int
virDomainGroupAccumulateCPUs(virDomainGroupDefPtr dgrp,
                             bool claim,
                             bool refresh)
{
    int ret = -1;
    size_t cnt = dgrp->pcpus;
    virDomainGroupGuestsPtr *dguests = dgrp->guest;
    virBitmapPtr ovcommitmask = NULL;
    virBitmapPtr cpuguestmask = dgrp->dgrpmask;

    /* If there is no formerly arranged domain group mask, there
     * was likely manual tweaking of guest domains while guests
     * were in execution. */
    if (!cpuguestmask)
        goto error;

    if (!(ovcommitmask = virBitmapNewCopy(dgrp->pcpumask)))
        goto error;

    if (virBitmapCopy(cpuguestmask, dgrp->pcpumask) < 0)
        goto error;

    if (refresh) {
        dgrp->DG_pacnt = dgrp->pcpus * dgrp->oscnt;
        dgrp->DG_vrcnt = 0;
    }

    while (cnt--) {
        size_t i;
        virDomainGroupGuestsPtr dguest = *dguests++;
        virDomainDefPtr def;

        if (!dguest)
            continue;

        def = dguest->def;
        if (!def)
            continue;

        for (i = 0; i < virDomainDefGetVcpusMax(def); i++) {
            virDomainVcpuDefPtr vcpu = virDomainDefGetVcpu(def, i);

            if (!vcpu->online)
                continue;

            if (vcpu->cpupack) {
                ssize_t pcpu = -1;

                while ((pcpu = virBitmapNextSetBit(vcpu->cpupack, pcpu)) >= 0) {
                    if (claim) {
                        ignore_value(virBitmapSetBit(cpuguestmask, pcpu));
                    } else {
                        virBitmapPtr target = cpuguestmask;

                        if (refresh) {
                            dgrp->DG_pacnt--;
                            dgrp->DG_vrcnt++;
                            if (!virBitmapIsBitSet(target, pcpu))
                                target = ovcommitmask;
                        }
                        ignore_value(virBitmapClearBit(target, pcpu));
                    }
                }

                /* If we're returning to the cpuguestmask pool
                 * then we should free our cpupack record too. */
                if (claim) {
                    virBitmapFree(vcpu->cpupack);
                    vcpu->cpupack = NULL;
                }
            }
        }
    }

    dgrp->DG_oversubscribed = false;
    if (dgrp->DG_vrcnt > dgrp->pcpus) {
        dgrp->DG_oversubscribed = true;
        if (refresh) {
            if (virBitmapCopy(cpuguestmask, ovcommitmask) < 0)
                goto error;
        }
    }

    ret = 0;

 error:
    virBitmapFree(ovcommitmask);
    return ret;
}


/*
 * Unsubscribe an active Guest Domain to the Domain Group.
 * Later this should become a chain'ed list.  For now
 * take the fast flat earth approach.
 *
 * WTH_XXX: RESOLVE.
 * Have the flat earth approach translated to a chained
 * set of dguest structures hanging under dgrp->guests.
 */
int
virDomainGroupUnsubscribeDomain(virDomainDefPtr def,
                                bool active,
                                virDomainGroupDefPtr dgrp)
{
    int ret = -1;
    size_t cnt = dgrp->pcpus;
    virDomainGroupGuestsPtr *dguests = dgrp->guest;

    while (dguests && cnt--) {
        virDomainGroupGuestsPtr dguest = *dguests;

        if (dguest && STREQ(dguest->name, def->name)) {

            if (active) {
                dguest->def = NULL;
                return 0;
            }

            if (dguest->def) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Cannot delete active domain "
                                 "'%s' from domain group '%s'"),
                               def->name, dgrp->name);
                goto error;
            }

            VIR_FREE(dguest->name);

            VIR_FREE(dguest);
            *dguests = dguest;

            dgrp->DG_refcnt--;
            ret = 0;
            break;
        }
        dguests++;
    }

    /* WTH_XXX: RESOLVE.
     * We need to add a SANITY check that the domain name
     * given does not still exists under the dgrp->guests
     * chain.  It is not possible to insert but imagine
     * the odd possibility something sneaks in.
     */

 error:
    return ret;
}


/*
 * Subscribe a Guest Domain to the Domain Group.
 * Later this should become a chain'ed list.  For now
 * take the fast flat earth approach.
 *
 * WTH_XXX: RESOLVE.
 * Have the flat earth approach translated to a chained
 * set of dguest structures hanging under dgrp->guests.
 */
int
virDomainGroupSubscribeDomain(virDomainDefPtr def,
                              bool active,
                              virDomainGroupDefPtr dgrp,
                              virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    size_t cnt = dgrp->pcpus;
    virDomainGroupGuestsPtr *dguests;

    if (!dgrp->guest &&
        VIR_ALLOC_N(dgrp->guest, dgbase->dgroup_guestpcpus) < 0)
        goto error;

    dguests = dgrp->guest;
    if (!dguests)
        return -1;

    while (*dguests && cnt--) {
        virDomainGroupGuestsPtr dguest = *dguests++;

        if (dguest &&
            STREQ(dguest->name, def->name)) {

            if (active) {
                dguest->def = def;
                return 0;
            }
        }
    }

    if (cnt) {
        virDomainGroupGuestsPtr dguest;

        if (VIR_ALLOC(dguest) < 0)
            goto error;

        if (VIR_STRDUP(dguest->name, def->name) < 0)
            goto error;

        if (active)
            dguest->def = def;
        else
            dgrp->DG_refcnt++;

        def->dgrp = dgrp;

        ret = 0;

        *dguests = dguest;
    }

 error:
    return ret;
}


static size_t
virDomainGroupDefGroupCnt(virDomainGroupDefPtr *dgrp)
{
    size_t ngrps = 0;

    while (dgrp && *dgrp) {
        virDomainGroupDefPtr s = *dgrp++;

        if (s->name)
            ngrps++;
    }

    return ngrps;
}


void
virDomainGroupDefFree(virDomainGroupDefPtr dgrp)
{
    if (!dgrp)
        return;

    VIR_FREE(dgrp->name);
    dgrp->name = NULL;

    VIR_FREE(dgrp);
    dgrp = NULL;

    return;
}


void
virDomainGroupDefArrayFree(virDomainGroupDefPtr *dgrp)
{
    virDomainGroupDefPtr *curr = dgrp;

    if (!dgrp)
        return;

    while (*curr)
        virDomainGroupDefFree(*curr++);

    VIR_FREE(dgrp);
    dgrp = NULL;

    return;
}


virDomainGroupDefPtr
virDomainGroupMatch(const char *group,
                    virDomainGroupDefPtr *dgroups,
                    virDomainGroupBasePtr dgbase)
{
    virDomainGroupDefPtr dgrp = NULL;
    virDomainGroupDefPtr *curr;

    if (!group)
        return NULL;

    if (!dgbase)
        return NULL;

    if (!dgbase->dgroups)
        return NULL;

    curr = dgroups ? dgroups : dgbase->dgroups;
    while (*curr) {
        virDomainGroupDefPtr s = *curr;

        if (s->name && (STREQ(s->name, group))) {
            dgrp = *curr;
            break;
        }
        curr++;
    }

    return dgrp;
}


virDomainGroupDefPtr
virDomainGroupRefresh(virDomainDefPtr def,
                      virDomainGroupBasePtr dgbase)
{
    virDomainGroupDefPtr dgrp;

    if (!def->dgroupname)
        return NULL;

    dgrp = virDomainGroupMatch(def->dgroupname, NULL, dgbase);
    if (!dgrp || !dgrp->name) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to find guest domain "
                         "'%s' within domain group '%s'"),
                       def->name, def->dgroupname);
        goto error;

    } else {
        if (STRNEQ(dgrp->name, def->dgroupname)) {
            VIR_FREE(def->dgroupname);

            if (VIR_STRDUP(def->dgroupname, dgrp->name) < 0)
                goto error;
        }
    }

    return dgrp;

 error:

    return NULL;
}


static bool
virDomainGroupvNUMAnodeAlign(size_t unit,
                             virDomainGroupBasePtr dgbase)
{
    bool ret;
    virCapsHostPtr host = dgbase->host;
    size_t align = (host->nnumaCell * host->cpu->threads);

    /* There should at least be one core per NUMA node */
    if (unit < align)
        return false;

    if (dgbase->dflags & VIR_DG_FLAGS_NUMA_ALIGN)
        align = host->cpu->threads;

    ret = (unit % align) ? false : true;

    return ret;
}



/**
 * virDomainGroupValidate
 *
 * Validate if given DGs match the host capabilities.
 *
 * dgbase - Domain Groups base.
 * refcnt - Track the Host offset pCPUs.
 * name   - Name of the domain group being validated, NULL if no particular
 *          domain group triggered the validation (e.g. during initial
 *          configuration).
 * pcpus  - The number of pCPUs for the domain group being validated, -1
 *          if not triggered for a particular domain group.
 */
static int
virDomainGroupValidate(virDomainGroupBasePtr dgbase,
                       int refcnt,
                       const char *name,
                       size_t pcpus)
{
    int dgpcpucnt = refcnt;
    virDomainGroupDefPtr *curr = dgbase->dgroups;

    /* Traverse the Domain Groups provided list */
    while (curr && *curr) {
        virDomainGroupDefPtr s = *curr++;

        if (!s->name)
            continue;

        /* An administrated DG should always fit the
         * minimum of 1 core per vNUMA node. */
        if (!virDomainGroupvNUMAnodeAlign(s->pcpus, dgbase)) {
            virReportError(VIR_ERR_PARSE_FAILED,
                           _("Cannot set domain group '%s' to %zu"
                             " pCPUs because that doesn't conform"
                             " to the host's NUMA topology"),
                           name ? name : s->name,
                           pcpus != -1 ? pcpus : s->pcpus);
            return -1;
        }

        dgpcpucnt += s->pcpus;

        /* Check if our 'host' pCPU set got depleted? */
        if (dgpcpucnt > dgbase->dgroup_guestpcpus) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Failed to set domain group '%s' to %zu pCPUs "
                             "because the number of allocated host pCPUs "
                             "would exceed the number of available host pCPUs, %zu"),
                           name ? name : s->name,
                           pcpus != -1 ? pcpus : s->pcpus,
                           dgbase->dgroup_guestpcpus);
            return -1;
        }
    }

    return dgpcpucnt;
}


/* Parse the XML definition for a group
 *
 * group has the form of
 *   <group name='coke' pcpus='4'/>
 */
static int
virDomainGroupNamePcpusDefParseXML(virDomainGroupDefPtr dgrp,
                                   xmlNodePtr node)
{
    unsigned int pcpus;
    VIR_AUTOFREE(char *) tmp = NULL;

    if (!(dgrp->name = virXMLPropString(node, "name"))) {
        virReportError(VIR_ERR_XML_ERROR, "%s",
                       _("Domain group property 'name' is missing from XML"));
        return -1;
    }

    if (!(tmp = virXMLPropString(node, "pcpus"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Domain group '%s' is missing property 'pcpus'"),
                       dgrp->name);
        return -1;
    }

    if (virStrToLong_uip(tmp, NULL, 10, &pcpus) < 0) {
        virReportError(VIR_ERR_XML_ERROR,
                       _("Domain group '%s' has an invalid 'pcpus' value: '%s'"),
                       dgrp->name, tmp);
        return -1;
    }
    VIR_FREE(tmp);

    if (!dgrp->oscnt)
        dgrp->oscnt = VIR_DG_OSCNT;

    dgrp->pcpus = pcpus;
    dgrp->DG_id = -1;
    dgrp->DG_pacnt = pcpus * dgrp->oscnt;

    return 0;
}


static virDomainGroupDefPtr
virDomainGroupDefParseXML(xmlXPathContextPtr ctxt,
                          virDomainGroupDefPtr *dgroups)
{
    int n;
    size_t i;
    virDomainGroupDefPtr dgrp = NULL;
    VIR_AUTOFREE(xmlNodePtr *) nodes = NULL;

    if ((n = virXPathNodeSet("./dgroup", ctxt, &nodes)) < 0)
        goto error;

    for (i = 0; i < n; i++) {
        virDomainGroupDefPtr *curr = dgroups;

        if (VIR_ALLOC(dgrp) < 0)
            return NULL;

        if (virDomainGroupNamePcpusDefParseXML(dgrp, nodes[i]))
            goto error;

        while (*curr) {
            if (STREQ(curr[0]->name, dgrp->name)) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Domain group '%s' is a duplicate"),
                               dgrp->name);
                goto error;
            }
            curr++;
        }
        *curr = dgrp;

        VIR_WARN("name=%s, pcpus=%zu", dgrp->name, dgrp->pcpus);
    }
    VIR_FREE(nodes);

 cleanup:
    return dgrp;

 error:
    virDomainGroupDefFree(dgrp);
    dgrp = NULL;
    goto cleanup;
}


virDomainGroupDefPtr
virDomainGroupDefParseNode(xmlDocPtr xml,
                           xmlNodePtr root,
                           virDomainGroupDefPtr *dgroups)
{
    xmlXPathContextPtr ctxt = NULL;
    virDomainGroupDefPtr ret = NULL;

    if (!virXMLNodeNameEqual(root, "domain")) {
        virReportError(VIR_ERR_XML_ERROR,
                       _("Discovered unexpected root element <%s>,"
                         " was expecting <domain>"),
                       root->name);
        goto cleanup;
    }

    ctxt = xmlXPathNewContext(xml);
    if (ctxt == NULL) {
        virReportOOMError();
        goto cleanup;
    }

    ctxt->node = root;

    ret = virDomainGroupDefParseXML(ctxt, dgroups);

 cleanup:
    xmlXPathFreeContext(ctxt);
    return ret;
}


static virDomainGroupDefPtr
virDomainGroupDefParse(const char *xmlStr,
                       const char *filename,
                       virDomainGroupDefPtr *dgroups)
{
    virDomainGroupDefPtr dgrp = NULL;
    xmlDocPtr xml;

    if ((xml = virXMLParse(filename, xmlStr, _("(domaingroup_definition)")))) {
        dgrp = virDomainGroupDefParseNode(xml, xmlDocGetRootElement(xml), dgroups);
        xmlFreeDoc(xml);
    }

    return dgrp;
}


static virDomainGroupDefPtr
virDomainGroupDefParseString(const char *xmlStr,
                             virDomainGroupDefPtr *dgroups)
{
    return virDomainGroupDefParse(xmlStr, NULL, dgroups);
}


int
virDomainGroupEnable(virDomainGroupBasePtr dgbase)
{
    if (dgbase)
        dgbase->enabled = true;

    return 0;
}


int
virDomainGroupAdd(const char *dgroupname,
                  const char *name,
                  virDomainDefPtr def,
                  virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr dgrp;

    if (!def)
        return 0;

    /* Fail if there is already a domain group
     * administration under a guest domain */
    if (dgbase->enabled && def->dgroupname) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       _("Cannot add domain '%s' to domain group '%s' because"
                         " it already belongs to domain group '%s'"),
                       name, dgroupname, def->dgroupname);
        goto error;
    }

    /* Escape if the domain group dgroupname does not exist */
    if (!(dgrp = virDomainGroupMatch(dgroupname, NULL, dgbase))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot add domain '%s' to non-existent"
                         " domain group '%s'"),
                       name, dgroupname);
        goto error;
    }

    ret = virDomainGroupSubscribeDomain(def, false, dgrp, dgbase);

 error:
    return ret;
}


int
virDomainGroupDelete(virDomainDefPtr def,
                     virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr dgrp;

    /* Fail if there is no domain group
     * administration under a guest domain */
    if (!def->dgroupname) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       _("Domain '%s' does not belong to a domain group"),
                       def->name);
        goto error;
    }

    /* Escape if the DG to delete from * does not exists */
    if (!(dgrp = virDomainGroupMatch(def->dgroupname, NULL, dgbase))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot delete domain '%s' from a non-existent"
                         " domain group"),
                       def->name);
        goto error;
    }

    ret = virDomainGroupUnsubscribeDomain(def, false, dgrp);

 error:
    return ret;
}


/* virDomainGroupCheckFit
 * The arranged DG pCPUs (npcpus) should always fit the amount of 'host'
 * available guest cpus (dgbase->dgroup_guestpcpus). */
static int
virDomainGroupCheckFit(virDomainGroupBasePtr dgbase,
                       unsigned int pcpus,
                       int refcnt,
                       const char *name,
                       unsigned int npcpus)
{
    int dgused_pcpus;

    if ((dgused_pcpus = virDomainGroupValidate(dgbase, refcnt, name, npcpus)) < 0)
        return -1;

    if ((dgused_pcpus - pcpus + npcpus) > dgbase->dgroup_guestpcpus) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Failed to set domain group '%s' to %u pCPUs because "
                         "that doesn't fit the number of host available CPUs"),
                       name, npcpus);
        return -1;
    }

    return 0;
}


int
virDomainGroupDefine(const char *xml,
                     virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr *src, *dst;
    virDomainGroupDefPtr *base = NULL, *dgroups = NULL;

    if (!dgbase)
        return 0;

    if (VIR_ALLOC_N(base, dgbase->dgroup_guestpcpus) < 0)
        goto error;

    /* Parse the XML input. Duplicates will expose! */
    if (!virDomainGroupDefParseString(xml, base))
        goto error;

    /* If Domain Groups are not administrated this is an initial
     * definition request.  Let 'base' be our dgroups reference! */
    if (dgbase->dgroups == NULL) {
        dgbase->dgroups = base;

        /* Are we over-running certain resource/criteria? */
        src = dgbase->dgroups;
        while (src && *src) {
            virDomainGroupDefPtr s = *src++;

            /* Re-apply cross list to fully match that what is provided. */
            if (virDomainGroupSetPcpus(s->name, s->pcpus, dgbase) < 0) {
                dgbase->dgroups = NULL;
                goto error;
            }
        }
        return 0;
    }

    /* Allocate a Domain Groups array for a work set */
    if (VIR_ALLOC_N(dgroups, dgbase->dgroup_guestpcpus) < 0)
        goto error;

    /* Set the list pointers */
    src = dgbase->dgroups;
    dst = dgroups;

    /* Copy the system Domain Groups array into our dgroup list */
    while (*src) {
        virDomainGroupDefPtr d, s = *src++;

        /* Duplicate the origine dgroup */
        if (VIR_ALLOC(d) < 0)
            goto error;

        if (VIR_STRDUP(d->name, s->name) < 0) {
            VIR_FREE(d);
            goto error;
        }

        d->pcpus = s->pcpus;
        d->oscnt = s->oscnt;

        /* Next */
        *dst++ = d;
    }

    /* Match our new 'define' input set and insert the elements
     * under the copy Domain Groups array 'dst' when unique */
    src = base;
    while (*src) {
        int refcnt = 0;
        virDomainGroupDefPtr d, s = *src++;

        /* Escape if this entry already exists */
        if (virDomainGroupMatch(s->name, dgroups, dgbase)) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Domain group '%s' already exists"),
                           s->name);
            goto error;
        }

        /* Escape if this entry fails the host NUMA and CPU Topology */
        if (!virDomainGroupvNUMAnodeAlign(s->pcpus, dgbase)) {
            virReportError(VIR_ERR_PARSE_FAILED,
                           _("Cannot set domain group '%s' to %zu pCPUs "
                             "because that doesn't conform to the host's"
                             " NUMA topology"),
                           s->name, s->pcpus);
            goto error;
        }

        /* Check if we're not running over host provided resources */
        refcnt += s->pcpus;
        if (virDomainGroupCheckFit(dgbase, s->pcpus, refcnt, s->name, s->pcpus) < 0)
            goto error;

        /* This entry is unique and good to go */
        if (VIR_ALLOC(d) < 0)
            goto error;

        if (VIR_STRDUP(d->name, s->name) < 0) {
            VIR_FREE(d);
            goto error;
        }

        d->pcpus = s->pcpus;
        d->oscnt = s->oscnt;

        /* Next */
        *dst++ = d;
    }

    /* The input 'base' Domain Group array 'dgroups' holds a
     * unique set that can be inserted under the system Domain
     * Groups array. */
    src = base;
    dst = dgbase->dgroups;
    while (*src) {
        virDomainGroupDefPtr d = *dst, s = *src;

        /* Are we appending to the array? */
        if (!d) {
           if (VIR_ALLOC(d) < 0)
               goto error;
        }

        /* A slot without a 'name' is a free slot */
        if (!d->name) {

            if (VIR_STRDUP(d->name, s->name) < 0)
                goto error;

            d->pcpus = s->pcpus;
            d->oscnt = s->oscnt;
            d->DG_id = -1;
            d->DG_pacnt = d->pcpus * d->oscnt;

            /* Add the newly defined Domain Group */
            *dst = d;

            src++;
        }
        dst++;
    }

    ret = 0;

 error:
    virDomainGroupDefArrayFree(base);
    virDomainGroupDefArrayFree(dgroups);

    return ret;
}


int
virDomainGroupUndefine(const char *name,
                       virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr dgrp;

    /* Escape if this entry already exists */
    if (!(dgrp = virDomainGroupMatch(name, NULL, dgbase))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot undefine non-existent domain group '%s'"),
                       name);
        goto error;
    }

    if (dgrp->DG_refcnt) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot undefine domain group '%s' because it"
                         " has %d domains associated with it"),
                       name, dgrp->DG_refcnt);
        goto error;
    }

    VIR_FREE(dgrp->name);
    memset(dgrp, 0, sizeof(*dgrp));

    ret = 0;

 error:

    return ret;
}


int
virDomainGroupRename(const char *name,
                     const char *newname,
                     virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr dgrp;

    /* Escape if the 'newname' entry already exists */
    if ((dgrp = virDomainGroupMatch(newname, NULL, dgbase))) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Cannot rename domain group '%s' to '%s' "
                         "because that domain group already exists"),
                       name, newname);
        goto error;
    }

    /* Escape if our domain group 'name' does not exists */
    if (!(dgrp = virDomainGroupMatch(name, NULL, dgbase))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot rename non-existent domain group '%s'"),
                       name);
        goto error;
    }

    VIR_FREE(dgrp->name);
    if (VIR_STRDUP(dgrp->name, newname) < 0)
        goto error;

    ret = dgrp->pcpus;

 error:

    return ret;
}


int
virDomainGroupGetpCPUs(const char *name,
                       virDomainGroupBasePtr dgbase,
                       virBitmapPtr target,
                       int type)
{
    int ret = -1;
    size_t ncpus = 0;
    ssize_t pcpu = -1;
    virDomainGroupDefPtr dgrp;

    if (!(dgrp = virDomainGroupMatch(name, NULL, dgbase)))
        goto cleanup;

    ncpus = dgrp->pcpus;
    switch (type) {

        case VIR_DOMAINGROUP_PCPU_ALL:
            if (dgrp->pcpumask) {
                if (virBitmapCopy(target, dgrp->pcpumask) < 0)
                    goto cleanup;
            }
            break;

        case VIR_DOMAINGROUP_PCPU_OCCUPIED:
            if (dgrp->DG_oversubscribed) {
                if (dgrp->pcpumask) {
                    if (virBitmapCopy(target, dgrp->pcpumask) < 0)
                        goto cleanup;
                }
            } else {
                if (dgrp->pcpumask && dgrp->dgrpmask) {

                    if (virBitmapCopy(target, dgrp->pcpumask) < 0)
                        goto cleanup;

                    while ((pcpu = virBitmapNextSetBit(dgrp->dgrpmask, pcpu)) >= 0)
                        ignore_value(virBitmapClearBit(target, pcpu));
                }
            }
            break;

        case VIR_DOMAINGROUP_PCPU_FREE:
            if (dgrp->dgrpmask) {
                if (virBitmapCopy(target, dgrp->dgrpmask) < 0)
                    goto cleanup;
            }
            break;
    }

    /* Count the bits set as they indicate the count of pCPUs */
    if (!ncpus) {
        while ((pcpu = virBitmapNextSetBit(target, pcpu)) >= 0)
            ncpus++;
    }
    ret = ncpus;

 cleanup:

    return ret;
}


int
virDomainGroupSetPcpus(const char *name,
                       unsigned int npcpus,
                       virDomainGroupBasePtr dgbase)
{
    virDomainGroupDefPtr dgrp;

    dgrp = virDomainGroupMatch(name, NULL, dgbase);
    if (!dgrp) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Domain group '%s' does not exist"),
                       name);
        return -1;
    }

    /* The arranged DG pCPUs should always align */
    if (!virDomainGroupvNUMAnodeAlign(npcpus, dgbase)) {
        virReportError(VIR_ERR_PARSE_FAILED,
                       _("Cannot set domain group '%s' to %u pCPUs "
                         "because that doesn't conform to the host's"
                         " NUMA topology"),
                       name, npcpus);
        return -1;
    }

    /* The arranged DG pCPUs * oversubscription (2) should
     * always fit the amount of vCPUs in effect. */
    if ((npcpus * dgrp->oscnt) < dgrp->DG_vrcnt) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Failed to set domain group '%s' to %u pCPUs because"
                         " that doesn't fit the number of active vCPUs, %d"),
                       name, npcpus, dgrp->DG_vrcnt);
        return -1;
    }

    /* Is this a valid operation at all? */
    if (virDomainGroupCheckFit(dgbase, dgrp->pcpus, 0, name, npcpus) < 0)
        return -1;

    /* Is there any change? */
    if (npcpus == dgrp->pcpus)
        return 0;

    /* WARN report that we're transferring our DG->pcpus
     * to npcpus */
    VIR_WARN("name=%s, pcpus=%zu -> pcpus=%u",
            dgrp->name, dgrp->pcpus, npcpus);

    /* Apply new statistics to this Domain Group */
    dgrp->pcpus = npcpus;
    dgrp->DG_pacnt = dgrp->pcpus * dgrp->oscnt;

    /* Transfer the vCPUs back from the DG under the
     * host cpuguestmask */
    if (dgrp->DG_usecnt && dgrp->DG_prsvd) {
        ssize_t pcpu = -1;

        /* Seek and reclaim this Domain Group pinned
         * vcpus and return them into the DG pool as
         * we're about to recompute and repin whole */
        if (virDomainGroupAccumulateCPUs(dgrp, true, false) < 0)
            return -1;

        /* New reservations are forth coming */
        dgrp->DG_prsvd = 0;
        dgrp->DG_vrcnt = 0;

        while ((pcpu = virBitmapNextSetBit(dgrp->pcpumask, pcpu)) >= 0) {

            if (!virBitmapIsBitSet(dgrp->dgrpmask, pcpu)) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Failed to reclaim all domain vCPUs from "
                                 "domain group '%s' (missing %zu)"),
                               dgrp->name, pcpu);
                return -1;
            }

            ignore_value(virBitmapClearBit(dgrp->pcpumask, pcpu));
            ignore_value(virBitmapSetBit(dgbase->cpuguestmask, pcpu));
        }

        /* Clear former active group mask */
        virBitmapFree(dgrp->dgrpmask);
        dgrp->dgrpmask = NULL;

        if (virDomainGroupBasePin(dgbase) < 0)
            return -1;
    }

    return dgrp->pcpus;
}


int
virDomainGroupGetVcpus(const char *name,
                       virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr dgrp;

    if ((dgrp = virDomainGroupMatch(name, NULL, dgbase)))
        ret = dgrp->DG_vrcnt;

    return ret;
}


virDomaingroupPtr
virDomainGroupLookupByDomain(virConnectPtr conn,
                             virDomainDefPtr def,
                             virDomainDefPtr persistentDef,
                             virDomainGroupBasePtr dgbase)
{
    virDomainGroupDefPtr *dgrp = dgbase ? dgbase->dgroups : NULL;
    virDomainDefPtr target = def ? def : persistentDef;

    if (dgrp &&
        target->dgroupname) {

        while (*dgrp) {
            virDomainGroupDefPtr s = *dgrp++;

            if (s->name && (STREQ(s->name, target->dgroupname))) {
                const unsigned char uuid0[] = "exadata-dgroups";

                return virGetDomaingroup(conn, s->name, uuid0, s->DG_id);
            }
        }
    }

    return NULL;
}


int
virDomainGroupListAll(virConnectPtr conn,
                      virDomaingroupPtr **grps,
                      virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    virDomaingroupPtr *list;
    virDomainGroupDefPtr *dgrp = dgbase ? dgbase->dgroups : NULL;
    size_t ngrps;

    if (!dgrp)
        return 0;

    /* Are there any Domain Groups to list? */
    if ((ngrps = virDomainGroupDefGroupCnt(dgrp)) == 0)
        return 0;

    if (VIR_ALLOC_N(list, ngrps) < 0)
        goto cleanup;

    *grps = list;

    while (*dgrp) {
        virDomainGroupDefPtr s = *dgrp++;

        if (s->name) {
            const unsigned char uuid0[] = "exadata-dgroups";
            *list++ = virGetDomaingroup(conn, s->name, uuid0, s->DG_id);
        }
    }

    ret = ngrps;

 cleanup:

    return ret;
}


int
virDomainGroupInit(const char *configDir,
                   bool force,
                   bool dgroupsAlignNUMA,
                   virCapsHostPtr host,
                   size_t guestCPUs,
                   virBitmapPtr cpuguestmask,
                   virDomainGroupBasePtr *ref_dgbase)
{
    int ret = -1;
    virDomainGroupDefPtr *dgroups = NULL;
    virDomainGroupBasePtr dgbase = NULL;
    VIR_AUTOFREE(char *) domainDir = NULL;
    VIR_AUTOFREE(char *) domaingroupFile = NULL;

    /* If we're already initialized return an error */
    if (*ref_dgbase)
        return ret;

    if ((domaingroupFile =
        virDomainConfigFile(configDir, "domain/dgroups")) == NULL)
        return ret;

    /* NO "File"? ... Create the Domain Group infra structure */
    if (access(domaingroupFile, R_OK) < 0) {

        if (!force)
            return 0;

        if (virAsprintf(&domainDir, "%s/domain", configDir) < 0)
            return ret;

        if (virFileMakePath(domainDir) < 0) {
            virReportSystemError(errno,
                                 _("cannot create Domain Groups config"
                                   " directory '%s'"),
                                 domainDir);
            return ret;
        }
    }

    /* Allocate our Domain Groups data structure */
    if (VIR_ALLOC(dgbase) < 0)
        goto cleanup;

    /* the amount of Domain Groups can never exceed the host
     * carried dgbase->dgroup_guestpcpus.  */
    dgbase->dgroup_guestpcpus = guestCPUs;

    if (VIR_ALLOC_N(dgroups, dgbase->dgroup_guestpcpus) < 0)
        goto cleanup;

    /* Domain Groups accessible set of Guest Domain pinnable pCPUs */
    dgbase->cpuguestmask = cpuguestmask;

    /* Add a handle for validation towards host capabilities */
    dgbase->host = host;

    /* Inform that unaligned domain.groups are accepted. */
    if (!dgroupsAlignNUMA) {
        dgbase->dflags = VIR_DG_FLAGS_NUMA_ALIGN;
        VIR_WARN("Domain Groups vNUMA node core alignment is disabled");
    }

    /* Import the Domain Groups */
    if (!virDomainGroupDefParse(NULL, domaingroupFile, dgroups)) {
        VIR_FREE(dgroups);
    } else {
        /* Assign our Domain Groups array */
        dgbase->dgroups = dgroups;

        /* Are we over-running certain resource/criteria? */
        if (virDomainGroupValidate(dgbase, 0, NULL, -1) < 0)
            goto cleanup;

        if (virDomainGroupDefSaveStatus(configDir, dgbase) < 0)
            goto cleanup;
    }

    *ref_dgbase = dgbase;

    /* Leave a message for systemctl -l status libvirtd.service */
    VIR_WARN("Domain Groups are active for %zu guest domain pCPUs",
            dgbase->dgroup_guestpcpus);

    ret = 0;

 cleanup:
    if (ret) {
        VIR_FREE(dgroups);
        VIR_FREE(dgbase);
    }

    return ret;
}


int
virDomainGroupSaveXML(const char *configDir,
                      const char *xml)
{
    const char *DG_EMPTY = "<domain>\n</domain>\n";
    VIR_AUTOFREE(char *) configFile = NULL;

    if (!configDir)
        return 0;

    if ((configFile =
        virDomainConfigFile(configDir, "domain/dgroups")) == NULL)
        goto error;

    if (virFileMakePath(configDir) < 0) {
        virReportSystemError(errno,
                _("cannot create config directory '%s'"),
                configDir);
        goto error;
    }

    if (xml && (STREQ(xml, DG_EMPTY)))
        ignore_value(unlink(configFile));
    else
        return virXMLSaveFile(configFile,
                virXMLPickShellSafeComment("domaingroups", "exadata"),
                "edit", xml);
 error:
    return -1;
}


int
virDomainGroupDefFormatBuf(virBufferPtr buf,
                           virDomainGroupDefPtr *dgrp)
{
    virBufferAddLit(buf, "<domain>\n");
    virBufferAdjustIndent(buf, 2);

    while (*dgrp) {
        virDomainGroupDefPtr s = *dgrp++;

        if (s->name && s->pcpus)
            virBufferAsprintf(buf, "<dgroup name='%s' pcpus='%zu'/>\n", s->name, s->pcpus);
    }

    virBufferAdjustIndent(buf, -2);
    virBufferAddLit(buf, "</domain>\n");

    return 0;
}


char *
virDomainGroupDefFormat(virDomainGroupDefPtr *dgrp)
{
    virBuffer buf = VIR_BUFFER_INITIALIZER;

    if (virDomainGroupDefFormatBuf(&buf, dgrp) < 0) {
        virBufferFreeAndReset(&buf);
        return NULL;
    }

    if (virBufferCheckError(&buf) < 0)
        return NULL;

    return virBufferContentAndReset(&buf);
}


int
virDomainGroupDefSaveStatus(const char *configDir,
                            virDomainGroupBasePtr dgbase)
{
    int ret = -1;
    char *xml = NULL;
    virDomainGroupDefPtr *dgrp = dgbase ? dgbase->dgroups : NULL;

    if (!dgrp)
        goto cleanup;

    if (!(xml = virDomainGroupDefFormat(dgrp)))
        goto cleanup;

    if (virDomainGroupSaveXML(configDir, xml) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    VIR_FREE(xml);

    return ret;
}

#endif /* defined(ENABLE_EXADATA) */
