/*
 * virdomaingroup.h: ExaData Domain Groups XML processing
 *
 * Copyright (c) 2019 Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "internal.h"
#include "viruuid.h"
#include "virpci.h"
#include "virxml.h"

#if defined(ENABLE_EXADATA)

# define VIR_DG_OSCNT 2         /* Domain Group oversubscribe factor */

typedef enum {
    VIR_DG_FLAGS_CURRENT    = 0,        /* Take current flags */
    VIR_DG_FLAGS_NUMA_ALIGN = 1 << 0,   /* Enforce vNUMA core/node alignment */
} virDomainGroupFlags;

/* Domain Group Guests in memory array */
typedef struct _virDomainGroupGuests virDomainGroupGuests;
typedef virDomainGroupGuests *virDomainGroupGuestsPtr;

struct _virDomainGroupGuests {
    char *name;
    virDomainDefPtr def;
};

/* Domain Group in memory array */
typedef struct _virDomainGroupDef virDomainGroupDef;
typedef virDomainGroupDef *virDomainGroupDefPtr;

struct _virDomainGroupDef {

    /* Advertised to /etc/libvirt/qemu/domain/groups.xml */
    char *name;                 /* Domain Group name */
    unsigned char uuid[VIR_UUID_BUFLEN]; /* Domain Group UUID */
    size_t pcpus;               /* Domain pCPU count */
    unsigned int oscnt;         /* Domain Group oversubscription count */

    /* Internal references */
    unsigned int DG_id;         /* Domain Group index ID */
    unsigned int DG_prsvd;      /* Domain Group pCPU reserved count */
    unsigned int DG_pacnt;      /* Domain Group pCPU available count */
    unsigned int DG_vrcnt;      /* Domain Group vCPU reference count */
    bool DG_oversubscribed;     /* Domain Group oversubscription in effective */

    int DG_refcnt;              /* Domain Group referenced by guest count */
    int DG_usecnt;              /* Domain Group in use by guest count */

    virBitmapPtr pcpumask;      /* Domain pCPU bitmask physical to dgroup */
    virBitmapPtr dgrpmask;      /* Domain vCPU:pCPU active mapped bitmask */

    virDomainGroupGuestsPtr *guest; /* Domain definition array towards guest running */
};

/* Domain Group base reference from capabilities */
typedef struct _virDomainGroupBase virDomainGroupBase;
typedef virDomainGroupBase *virDomainGroupBasePtr;

struct _virDomainGroupBase {

    bool enabled;               /* Domain Groups are enabled */
    unsigned int dflags;        /* Domain Groups exclusive flags */

    int dgroup_id;              /* Identifier */
    size_t dgroup_guestpcpus;   /* host pcpus */
    virBitmapPtr cpuguestmask;  /* host cpuguestmask */
    virCapsHostPtr host;        /* host capabilities */
    virDomainGroupDefPtr *dgroups; /* Domain Groups Array */
};


int
virDomainGroupBasePin(virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupAccumulateCPUs(virDomainGroupDefPtr dgrp,
                             bool claim,
                             bool refresh)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupUnsubscribeDomain(virDomainDefPtr def,
                                bool active,
                                virDomainGroupDefPtr dgrp)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupSubscribeDomain(virDomainDefPtr def,
                              bool active,
                              virDomainGroupDefPtr dgrp,
                              virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

void
virDomainGroupDefFree(virDomainGroupDefPtr dgrp);

void
virDomainGroupDefArrayFree(virDomainGroupDefPtr *dgrp);

virDomainGroupDefPtr
virDomainGroupMatch(const char *group,
                    virDomainGroupDefPtr *dgrp,
                    virDomainGroupBasePtr dgbase);

virDomainGroupDefPtr
virDomainGroupRefresh(virDomainDefPtr def,
                      virDomainGroupBasePtr dgbase);

virDomainGroupDefPtr
virDomainGroupDefParseNode(xmlDocPtr xml,
                           xmlNodePtr root,
                           virDomainGroupDefPtr *dgroups)
    ATTRIBUTE_NONNULL(3);


int
virDomainGroupInit(const char *configDir,
                   bool force,
                   bool dgroupsAlignNUMA,
                   virCapsHostPtr host,
                   size_t guestCPUs,
                   virBitmapPtr cpuguestmask,
                   virDomainGroupBasePtr *dgbase);

int
virDomainGroupSaveXML(const char *configDir,
                      const char *xml)
    ATTRIBUTE_NONNULL(2);

int
virDomainGroupEnable(virDomainGroupBasePtr dgbase);

int
virDomainGroupAdd(const char *dgroupname,
                  const char *name,
                  virDomainDefPtr def,
                  virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1) ATTRIBUTE_NONNULL(2);

int
virDomainGroupDelete(virDomainDefPtr def,
                     virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupDefine(const char *xml,
                     virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupUndefine(const char *name,
                       virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

int
virDomainGroupRename(const char *name,
                     const char *newname,
                     virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1) ATTRIBUTE_NONNULL(2);

int
virDomainGroupGetpCPUs(const char *name,
                       virDomainGroupBasePtr dgbase,
                       virBitmapPtr target,
                       int type)
    ATTRIBUTE_NONNULL(2) ATTRIBUTE_NONNULL(3);

int
virDomainGroupSetPcpus(const char *name,
                       unsigned int npcpus,
                       virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(2);

int
virDomainGroupGetVcpus(const char *name,
                       virDomainGroupBasePtr dgbase)
    ATTRIBUTE_NONNULL(1);

virDomaingroupPtr
virDomainGroupLookupByDomain(virConnectPtr conn,
                             virDomainDefPtr def,
                             virDomainDefPtr persistentDef,
                             virDomainGroupBasePtr dgbase);

int
virDomainGroupListAll(virConnectPtr conn,
                      virDomaingroupPtr **grps,
                      virDomainGroupBasePtr dgbase);

char *
virDomainGroupDefFormat(virDomainGroupDefPtr *def);

int
virDomainGroupDefFormatBuf(virBufferPtr buf,
                           virDomainGroupDefPtr *def)
    ATTRIBUTE_NONNULL(2);

int
virDomainGroupDefSaveStatus(const char *configDir,
                            virDomainGroupBasePtr dgbase);

#endif /* defined(ENABLE_EXADATA) */
