/*
 * qemu_hugepages.c: QEMU memory control
 *
 * Copyright (c) 2020 Oracle and/or its affiliates
 *
 * Autonomous memoryBacking nr_hugepage reservations under control of libvirt
 */

#include <config.h>

#include "viralloc.h"
#include "virlog.h"
#include "virnuma.h"
#include "virhostmem.h"

#include "qemu_conf.h"
#include "qemu_hugepages.h"
#include "qemu_process.h"

#define VIR_FROM_THIS VIR_FROM_QEMU

VIR_LOG_INIT("qemu.qemu_memory");

#if defined(ENABLE_EXADATA)
# define HP_TARGET (GuestIsNUMA ? cell : -1)

/* AHP_DEBUG: Autonomous Huge Pages instrumentation for reserving hugetlbfs managed pages */
# define LOG_VERBOSE 0
# define LOG_INSTRUMENTATION 1

# if LOG_VERBOSE
#  define DEB_VERBOSE(a, ...) { \
    fprintf(stderr, "%s; ", __FUNCTION__); \
    fprintf(stderr, a, __VA_ARGS__); \
    fprintf(stderr, "\n"); }
# else
#  if LOG_INSTRUMENTATION
#   define DEB_VERBOSE(a, ...) { \
        if (cfg->exadataConfig && \
            cfg->exadataConfig->AHP_debug) \
        VIR_WARN(a, __VA_ARGS__); }
#  else
#   define DEB_VERBOSE(a, ...)
#  endif
# endif

/* _qemuRetrySetPagePoolSize
 *
 * Controlling the Linux Page Pool Size to target 'host' arranged 2-MiB or 1-GiB
 * acquisition by virNumaSetPagePoolSize is surprising erratic when operated per SysFS
 * NUMA node-specific entries.  Requests are interrupted midstream and in need of retry
 * till the operation finishes.   Returning an error to its callee results in a fatality to
 * the arranged 'guest' domain toolstack, so this routine provides a mechanism to retry by
 * timeout, which propagates to the best operational demand.  Low demand steers towards
 * the minimum timeout, which high demand directs to a long timeout.  These timeouts,
 * if in effect, add at max a couple of seconds to a NUMA node memory acquisition
 * or release or for interactive individual DIMM HOT-PLUG and HOT-UNPLUG operations.
 * If an unusual event happens, we report the system error but never inform the callee
 * about the fatality. Hence there is _NO_ Errno anyways, as such would result in a full
 * libvirt stack caveat.
 *
 * WTH_RESOVE: This needs future attention; errors need to be dealt with.
 */
static void
_qemuRetrySetPagePoolSize(virQEMUDriverConfigPtr cfg,
                          virDomainDefPtr def,
                          size_t cell,
                          int target,
                          size_t Hugepagesize,
                          unsigned long long whence,
                          bool append)
{
    int err = -1;
    size_t retry = cfg->exadataConfig->retryHugePages;
    static size_t timeout;

    if (!timeout)
        timeout = cfg->exadataConfig->timeoutHugePages;
    else if (timeout > cfg->exadataConfig->timeoutHugePages)
        timeout -= cfg->exadataConfig->timeoutHugePages;

    do {
        if (retry < cfg->exadataConfig->retryHugePages)
            timeout += cfg->exadataConfig->timeoutHugePages;
        usleep(timeout);
    } while (retry-- && (err = virNumaSetPagePoolSize(target, Hugepagesize, whence, append)) < 0);

    /* There's a bad condition.  Make a libvirt report but keep it silent. */
    if (err)
        virReportError(VIR_ERR_SYSTEM_ERROR,
                _("Failed to reset the reservation "
                  "of %lld pages under node %zu for domain '%s'"),
                whence, cell, def->name);

    return;
}


static int
_qemuGetNUMAPagesInfo(int node,
                      qemuHostNUMACellPageInfoPtr *pageinfo,
                      size_t *npageinfo)
{
    int ret = -1;
    unsigned int *pages_size = NULL;
    unsigned long long *pages_avail = NULL;
    unsigned long long *pages_free = NULL;
    size_t npages, i;

    if (virNumaGetPages(node, &pages_size,
                &pages_avail, &pages_free, &npages) < 0)
        goto cleanup;

    if (VIR_ALLOC_N(*pageinfo, npages) < 0)
        goto cleanup;

    *npageinfo = npages;

    for (i = 0; i < npages; i++) {
        (*pageinfo)[i].size = pages_size[i];
        (*pageinfo)[i].avail = pages_avail[i];
        (*pageinfo)[i].free  = pages_free[i];
    }

    ret = 0;

 cleanup:
    VIR_FREE(pages_free);
    VIR_FREE(pages_avail);
    VIR_FREE(pages_size);
    return ret;
}


int
qemuDomainInInit(virQEMUDriverPtr driver,
                  virDomainDefPtr def,
                  virQEMUDriverConfigPtr cfg,
                  int state)
{
    int ret = 0;
    bool keep = true;
    static int _inpause;

    /* Only hold guest domains that require AHP acquisition */
    if (cfg->exadataConfig &&
        cfg->exadataConfig->enaAutonomousHugepages &&
        def->mem.nhugepages &&
        def->mem.hugepages && !def->mem.hugepages[0].nodemask) {

        do {
            switch (state) {
                case QEMU_AHP_ACQUIRE_SEM:
                    {
                        virMutexLock(&driver->lock);
                        if (_inpause == 0) {
                            _inpause = def->id;
                            ret = _inpause;
                            keep = false;
                        }
                        virMutexUnlock(&driver->lock);
                    }
                    break;

                case QEMU_AHP_ACQUIRE_SEM_LOCKED:
                    {
                        virMutexLock(&driver->lock);
                        if (_inpause == 0 ||
                            _inpause == def->id) {
                            _inpause = def->id;
                            ret = _inpause;
                            keep = false;
                            break;
                        }
                        virMutexUnlock(&driver->lock);
                    }
                    break;

                case QEMU_AHP_RELEASE_SEM:
                    {
                        virMutexLock(&driver->lock);
                        _inpause = 0;
                        ret = _inpause;
                        keep = false;
                        virMutexUnlock(&driver->lock);
                    }
                    break;

                case QEMU_AHP_RELEASE_SEM_LOCKED:
                    {
                        _inpause = 0;
                        ret = _inpause;
                        keep = false;
                        virMutexUnlock(&driver->lock);
                    }
                    break;

                default:
                    ret = -1;
                    keep = false;
                    break;
            }

            /* Relax this guest domain retrying */
            if (keep)
                usleep(cfg->exadataConfig->timeoutHugePages);

        } while (keep && ret >= 0);

        /* Below can not happen, I'd rather assert,
         * ... crawl into a dark corner and die */
        if (ret < 0)
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("AHP state yielded an error for request '%d'"),
                           state);
    }
    return ret;
}


int
qemuMemAcquireHostHugepages(virQEMUDriverPtr driver,
                            virDomainDefPtr def,
                            virQEMUDriverConfigPtr cfg,
                            unsigned long long DimmSize,
                            int targetNode)
{
    if (geteuid() == 0 &&
        cfg->exadataConfig &&
        cfg->exadataConfig->enaAutonomousHugepages &&
        cfg->exadataConfig->pctHugePages &&
        def->mem.nhugepages &&
        def->mem.hugepages && !def->mem.hugepages[0].nodemask) {

        static bool report = true;
        qemuHostNUMACellPageInfoPtr pageinfo = NULL;
        qemuGuestNUMACellReqPtr nodepages = NULL;
        virCapsPtr caps = driver->caps;
        virCapsHostPtr host = &caps->host;
        int err = -1;
        size_t startcell = 0, endcell;
        size_t cell, ncells = host->nnumaCell;
        size_t psindx;
        size_t npageinfo;
        size_t sparePages;
        size_t Hugepagesize = 2048L;
        bool GuestIsNUMA = true;
        unsigned long long GuestNodeSize = 0, GuestSize = 0, HostNodeFree = 0, RawhNodeFree = 0, HostFree = 0;
        unsigned long long ExceededNodeSize = 0;
        unsigned long long memorybytes;

        /* Calculate the maximum percentage setting for a guest's
         * memory relative to the host's total available memory. */
        if (virHostMemGetInfo(&memorybytes, NULL) < 0)
            goto cleanup;

        if (virDomainDefGetMemoryTotal(def) > \
            ((cfg->exadataConfig->pctHugePages * (memorybytes / 1024L)) / 100L)) {
            virReportError(VIR_ERR_OPERATION_INVALID,
                           _("Domain '%s' configured memoryBacking of %lld KiB, exceeds the host "
                             "%d%% restricted hugepage acquisition. Max allowed is %lld KiB."),
                           def->name, virDomainDefGetMemoryTotal(def),
                           cfg->exadataConfig->pctHugePages,
                           ((cfg->exadataConfig->pctHugePages * (memorybytes / 1024L)) / 100L));
            goto cleanup;
        }

        if (cfg->exadataConfig->spareNodeSize < MIN_SPARE_NODESIZE)
            cfg->exadataConfig->spareNodeSize = MIN_SPARE_NODESIZE;

        if (report) {
            VIR_WARN("Hugepage spare size is set to %zu GiB reserved spare pages per NUMA node",
                     cfg->exadataConfig->spareNodeSize);
            report = false;
        }

        /* Is this a HOT-PLUG? */
        if (targetNode >= 0) {

            /* Will it fit the available slots? If so, then don't cause
             * an exception.  The libvirt domain code takes a reversed
             * approach towards HOT-PLUG memory, checking its available
             * slots postmortem. */
            if (def->nmems >= def->mem.memory_slots) {
                err = 0;
                goto cleanup;
            }

            /* HOT-PLUG goes by single NUMA nodes. */
            ncells = 1;
            startcell = targetNode;
            endcell = targetNode + ncells;
        } else {

            /* Get the vNUMA node count. */
            ncells = virDomainNumaGetNodeCount(def->numa);

            /* If none, there's always a NUMA node0! */
            if (!ncells) {
                ncells = 1;
                GuestIsNUMA = false;
            }
            endcell = ncells;
        }

        if (VIR_ALLOC_N(nodepages, endcell) < 0)
            goto cleanup;

        /* If the guest configured a specific hugepage setting,
         * overrule the default 2-MiB setting. */
        if (def->mem.hugepages[0].size)
            Hugepagesize = def->mem.hugepages[0].size;

        DEB_VERBOSE(("AHP: (%s) Hugepagesize = %zu"), def->name, Hugepagesize);

        switch (Hugepagesize) {
            case 2048L:
                psindx = PAGE_2M;
                break;

            case 1048576:
                psindx = PAGE_1G;
                break;

            default:
                goto cleanup;
        }

        sparePages = (cfg->exadataConfig->spareNodeSize * 1024 * 1024) / Hugepagesize;

        DEB_VERBOSE(("AHP: (%s) sparePages = %zu"), def->name, sparePages);

        /* LOCK domain hugepage acquisition. */
        virMutexLock(&driver->lock);

        for (cell = startcell; cell < endcell; cell++) {

            /* How much does this guest(domain) need */
            if (GuestIsNUMA) {
                if (DimmSize)
                    GuestNodeSize = DimmSize / ncells;
                else
                    GuestNodeSize = virDomainNumaGetNodeMemorySize(def->numa, cell);
            } else {
                GuestNodeSize = virDomainDefGetMemoryTotal(def);
            }

            /* How much is available under the system.slice (host) */
            if (_qemuGetNUMAPagesInfo(HP_TARGET, &pageinfo, &npageinfo) < 0)
                goto error;

            /* Is there still sufficient space to build a sparePages buffer */
            RawhNodeFree = (4L * (pageinfo[PAGE_4K].free));
            RawhNodeFree = Hugepagesize * (RawhNodeFree / Hugepagesize);
            HostNodeFree = RawhNodeFree;

            DEB_VERBOSE(("AHP: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld [RawH %lld]"),
                    def->name, cell, GuestNodeSize, HostNodeFree, RawhNodeFree);

            if (HostNodeFree < (Hugepagesize * sparePages)) {
                if (cfg->exadataConfig->forceStrictNumaBinding &&
                   (HostNodeFree > Hugepagesize)) {
                    virReportError(VIR_ERR_OPERATION_INVALID,
                                   _("Insufficient memory to preserve host NUMA spare memory: "
                                     "node %zu spare pages, spare %lld, reserved %ld"),
                                   cell, HostNodeFree, (Hugepagesize * sparePages));
                    goto error;
                }
                HostNodeFree -= Hugepagesize;
            } else {
                HostNodeFree -= (Hugepagesize * sparePages);
            }

            /* Adjust the total HostFree count */
            HostFree += HostNodeFree;
            GuestSize += GuestNodeSize;

            DEB_VERBOSE(("+1a: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld [RawH %lld]"),
                    def->name, cell, GuestNodeSize, HostNodeFree, RawhNodeFree);

            if (GuestNodeSize > HostNodeFree) {
                if (cfg->exadataConfig->forceStrictNumaBinding) {
                    virReportError(VIR_ERR_OPERATION_INVALID,
                                   _("Insufficient memory available under NUMA "
                                     "node %zu, available %lld KiB, needed %lld KiB"),
                                   cell, HostNodeFree, GuestNodeSize);
                    goto error;
                }

                ExceededNodeSize += (GuestNodeSize - HostNodeFree);
                GuestNodeSize = HostNodeFree;
                nodepages[cell].occupied = true;

                DEB_VERBOSE(("+1b: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld, ExceededNodeSize %lld"),
                        def->name, cell, GuestNodeSize, HostNodeFree, ExceededNodeSize);
            }

            nodepages[cell].nr_avail = pageinfo[psindx].avail;
            nodepages[cell].nr_pages = GuestNodeSize;
            nodepages[cell].nr_free  = HostNodeFree;
        }

        /* Redistribute the ExceededNodeSize over the NUMA nodes
         * that have free space left.  First come, first served! */
        if (ExceededNodeSize) {
            for (cell = startcell; cell < endcell; cell++) {
                size_t i = endcell - cell - 1;

                /* If there's space under this cell */
                if (!nodepages[i].occupied) {
                    GuestNodeSize = nodepages[i].nr_pages;
                    HostNodeFree  = nodepages[i].nr_free;
                    HostNodeFree -= GuestNodeSize;

                    DEB_VERBOSE(("+2a: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld, ExceededNodeSize %lld"),
                            def->name, i, GuestNodeSize, HostNodeFree, ExceededNodeSize);

                    if (ExceededNodeSize > HostNodeFree) {
                        ExceededNodeSize -= HostNodeFree;
                        GuestNodeSize += HostNodeFree;
                    } else {
                        GuestNodeSize += ExceededNodeSize;
                        ExceededNodeSize = 0;
                    }

                    DEB_VERBOSE(("+2b: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld, ExceededNodeSize %lld"),
                            def->name, i, GuestNodeSize, HostNodeFree, ExceededNodeSize);

                    nodepages[i].nr_pages = GuestNodeSize;
                    nodepages[i].nr_free  = HostNodeFree;
                }
            }

            if (ExceededNodeSize) {
                virReportError(VIR_ERR_OPERATION_INVALID,
                               _("Insufficient memory available %lld KiB "
                                 "under the NUMA nodes to balance '%s' %lld KiB "
                                 "requested size.  Exceeds %lld KiB"),
                               HostFree, def->name, GuestSize, ExceededNodeSize);
                goto error;
            }
        }

        /* We still exist, so there's a fit! */
        for (cell = startcell; cell < endcell; cell++) {

            GuestNodeSize = nodepages[cell].nr_pages;
            HostNodeFree = nodepages[cell].nr_free;

            DEB_VERBOSE(("+3a: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld"),
                    def->name, cell, GuestNodeSize, HostNodeFree);

            if (virNumaSetPagePoolSize(HP_TARGET, Hugepagesize,
                        GuestNodeSize / Hugepagesize, true) < 0) {

                virReportError(VIR_ERR_OPERATION_INVALID,
                               _("Failed to acquire %lld required %zu-KiB "
                                 "pages under NUMA node %zu"),
                              GuestNodeSize / Hugepagesize, Hugepagesize, cell);

                /* Restore (reset) the last functional page sets. */
                do {
                    unsigned long long AvailNodeSize;

                    AvailNodeSize = nodepages[cell].nr_avail;
                    DEB_VERBOSE(("+3b: (%s)->cell %zu, GuestNodeSize %lld, HostNodeFree %lld AvailNodeSize %lld"),
                            def->name, cell, GuestNodeSize, HostNodeFree, AvailNodeSize);

                    _qemuRetrySetPagePoolSize(cfg, def, cell,
                            HP_TARGET, Hugepagesize, AvailNodeSize, false);

                } while (targetNode < 0 && cell--);

                goto error;
            }
        }

        /* Mark the hugepage acquisition perimeter. */
        def->acquiredHugepages = true;
        err = 0;

 error:
        /* UNLOCK domain hugepage acquisition. */
        virMutexUnlock(&driver->lock);

 cleanup:
        VIR_FREE(pageinfo);
        VIR_FREE(nodepages);

        return err;
    }

    return 0;
}


int
qemuMemReleaseHostHugepages(virQEMUDriverPtr driver,
                            virDomainDefPtr def,
                            virDomainMemoryDefPtr mem)
{
    VIR_AUTOUNREF(virQEMUDriverConfigPtr) cfg = virQEMUDriverGetConfig(driver);

    if (geteuid() == 0 &&
        cfg->exadataConfig &&
        cfg->exadataConfig->enaAutonomousHugepages &&
        def->mem.nhugepages &&
        def->mem.hugepages && !def->mem.hugepages[0].nodemask) {

        virCapsPtr caps = driver->caps;
        virCapsHostPtr host = &caps->host;
        int err = -1;
        bool GuestIsNUMA = true;
        size_t Hugepagesize = 2048L;
        size_t npageinfo;
        size_t psindx;
        size_t startcell = 0, endcell;
        size_t cell, ncells = host->nnumaCell;
        unsigned long long  acquiredHugepages = 0;
        qemuHostNUMACellPageInfoPtr pageinfo = NULL;

        /* If this is a HOT-UNPLUG assure it is within a
         * valid slot range. */
        if (mem && (def->nmems >= def->mem.memory_slots))
            return 0;

        if (!ncells) {
            /* There's always a NUMA node0. */
            ncells = 1;
            GuestIsNUMA = false;
        }
        endcell = ncells;

        /* If the guest configured a specific hugepage setting,
         * overrule the default 2-MiB setting. */
        if (def->mem.hugepages[0].size)
            Hugepagesize = def->mem.hugepages[0].size;

        DEB_VERBOSE(("AHP: (%s) Hugepagesize = %zu"), def->name, Hugepagesize);

        switch (Hugepagesize) {
            case 2048L:
                psindx = PAGE_2M;
                break;

            case 1048576:
                psindx = PAGE_1G;
                break;

            default:
                goto cleanup;
        }

        /* LOCK domain hugepage release. */
        if (qemuDomainInInit(driver, def, cfg, QEMU_AHP_ACQUIRE_SEM_LOCKED) < 0)
            goto cleanup;

        if (def->acquiredHugepages) {

            /* Is this an individual memory hot-unplug request? */
            if (mem) {
                cell = mem->targetNode;
                if (_qemuGetNUMAPagesInfo(cell, &pageinfo, &npageinfo) < 0)
                    goto error;

                acquiredHugepages = mem->size / Hugepagesize;

                _qemuRetrySetPagePoolSize(cfg, def, cell,
                        HP_TARGET, Hugepagesize, pageinfo[psindx].avail - acquiredHugepages, false);
            } else {

                DEB_VERBOSE(("AHP: (%s), GuestSize %lld"), def->name, virDomainDefGetMemoryTotal(def));

                /* Guest domain cleanup, tear down all vNUMA node acquired memory. */
                for (cell = startcell; cell < endcell; cell++) {

                    if (_qemuGetNUMAPagesInfo(HP_TARGET, &pageinfo, &npageinfo) < 0)
                        goto error;

                    acquiredHugepages = pageinfo[psindx].avail - pageinfo[psindx].free;

                    DEB_VERBOSE(("+1a: (%s)->cell %zu, releasing %lld acquiredHugepages"),
                            def->name, cell, pageinfo[psindx].free);

                    _qemuRetrySetPagePoolSize(cfg, def, cell,
                            HP_TARGET, Hugepagesize, acquiredHugepages, false);
                }
                def->acquiredHugepages = false;
            }
        }
        err = 0;

 error:
        /* UNLOCK domain hugepage release. */
        if (qemuDomainInInit(driver, def, cfg, QEMU_AHP_RELEASE_SEM_LOCKED) < 0)
            err = -1;

 cleanup:
        VIR_FREE(pageinfo);

        return err;
    }

    return 0;
}
#endif /* defined(ENABLE_EXADATA) */
